import pandas as pd
from typing import List, Tuple, Union, Optional
from chemprop import data, featurizers, models, nn
from lightning import pytorch as pl
from sklearn.preprocessing import StandardScaler
import torch
from torch.utils.data import DataLoader
import numpy as np
import multiprocessing
from tqdm import tqdm
import asyncio
from funcnodes_core.utils.functions import make_run_in_new_process


def make_model(scaler: StandardScaler) -> models.MPNN:
    mp = nn.BondMessagePassing()

    agg = nn.MeanAggregation()

    output_transform = nn.UnscaleTransform.from_standard_scaler(scaler)
    predictor = nn.RegressionFFN(output_transform=output_transform)
    batch_norm = True
    metric_list = [
        nn.metrics.RMSE(),
        nn.metrics.MAE(),
    ]  # Only the first metric is used for training and early stopping
    mpnn = models.MPNN(
        mp, agg, predictor=predictor, batch_norm=batch_norm, metrics=metric_list
    )
    return mpnn


def model_to_scaler(model: models.MPNN) -> StandardScaler:
    scaler = model.predictor.output_transform.to_standard_scaler()
    return scaler


async def make_data(
    df: pd.DataFrame,
    smiles_column: str,
    target_columns: Union[str, List[str]],
    split=(0.8, 0.1, 0.1),
    scaler: StandardScaler = None,
    split_type="random",
    progess_bar: Optional[tqdm] = None,
    in_new_thread: bool = True,
) -> Tuple[
    DataLoader,
    DataLoader,
    DataLoader,
    StandardScaler,
]:
    def _make_data(
        df, smiles_column, target_columns, split, scaler, split_type, progess_bar
    ):
        num_workers = 0
        if isinstance(target_columns, str):
            target_columns = [target_columns]
        if progess_bar is not None:
            progess_bar.set_description("Loading data")
            # set the total number of iterations to 3
            progess_bar.total = 4
            progess_bar.n = 0
            progess_bar.refresh()
        smis = df.loc[:, smiles_column].values
        ys = df.loc[:, target_columns].values
        all_data = [data.MoleculeDatapoint.from_smi(smi, y) for smi, y in zip(smis, ys)]
        if progess_bar is not None:
            progess_bar.update(1)
        mols = [
            d.mol for d in all_data
        ]  # RDkit Mol objects are use for structure based splits
        train_indices, val_indices, test_indices = data.make_split_indices(
            mols, split_type, split
        )
        train_data, val_data, test_data = data.split_data_by_indices(
            all_data, train_indices, val_indices, test_indices
        )
        if progess_bar is not None:
            progess_bar.update(1)
        featurizer = featurizers.SimpleMoleculeMolGraphFeaturizer()
        train_dset = data.MoleculeDataset(train_data[0], featurizer)
        scaler = train_dset.normalize_targets(scaler)

        val_dset = data.MoleculeDataset(val_data[0], featurizer)
        val_dset.normalize_targets(scaler)

        test_dset = data.MoleculeDataset(test_data[0], featurizer)

        if progess_bar is not None:
            progess_bar.update(1)

        train_loader = data.build_dataloader(train_dset, num_workers=num_workers)

        val_loader = data.build_dataloader(
            val_dset, num_workers=num_workers, shuffle=False
        )
        test_loader = data.build_dataloader(
            test_dset, num_workers=num_workers, shuffle=False
        )
        if progess_bar is not None:
            progess_bar.update(1)

        return (
            train_loader,
            val_loader,
            test_loader,
            scaler,
            train_indices[0],
            val_indices[0],
            test_indices[0],
        )

    if in_new_thread:
        # Offload the blocking _make_data function to a separate thread.
        return await asyncio.to_thread(
            _make_data,
            df,
            smiles_column,
            target_columns,
            split,
            scaler,
            split_type,
            progess_bar,
        )
    else:
        # Run _make_data directly in the current coroutine.
        return _make_data(
            df, smiles_column, target_columns, split, scaler, split_type, progess_bar
        )


class MetricCallback(pl.Callback):
    def __init__(self, queue):
        super().__init__()
        self.queue = queue

    def log_metrics2queue(self, trainer):
        metrics = trainer.callback_metrics
        # convert tensors to scalars
        metrics = {k: v.item() for k, v in metrics.items()}
        metrics["epoch"] = trainer.current_epoch
        self.queue.put(metrics)

    def on_train_epoch_end(
        self,
        trainer,
        pl_module,
    ):
        self.log_metrics2queue(trainer)

    def on_test_epoch_end(
        self, trainer: pl.Trainer, pl_module: pl.LightningModule
    ) -> None:
        self.log_metrics2queue(trainer)


def train(
    model: models.MPNN,
    train_loader: DataLoader,
    val_loader: DataLoader,
    test_loader: DataLoader,
    max_epochs=20,
    queue=None,
) -> Tuple[models.MPNN, dict]:
    callbacks = [MetricCallback(queue)] if queue else []
    trainer = pl.Trainer(
        logger=False,
        enable_progress_bar=True,
        accelerator="auto",
        max_epochs=max_epochs,  # number of epochs to train for
        callbacks=callbacks,
        enable_checkpointing=False,
        enable_model_summary=False,
    )

    trainer.fit(model, train_loader, val_loader)

    results = trainer.test(model, test_loader)
    return model, results


def _train_in_subprocess_entry(
    model,
    train_loader,
    val_loader,
    test_loader,
    max_epochs,
    return_dict,
    queue: multiprocessing.Queue,
):
    model, results = train(
        model,
        train_loader,
        val_loader,
        test_loader,
        max_epochs,
        queue,
    )
    return_dict["model"] = model
    return_dict["results"] = results
    return return_dict


_train_in_subprocess = make_run_in_new_process(_train_in_subprocess_entry)


async def train_in_subprocess(
    model: models.MPNN,
    train_loader: DataLoader,
    val_loader: DataLoader,
    test_loader: DataLoader,
    max_epochs=20,
) -> Tuple[models.MPNN, dict]:
    manager = multiprocessing.Manager()
    return_dict = manager.dict()
    queue = manager.Queue()

    # p = multiprocessing.Process(
    #     target=_train_in_subprocess_entry,
    #     args=(
    #         model,
    #         train_loader,
    #         val_loader,
    #         test_loader,
    #         max_epochs,
    #         return_dict,
    #         queue,
    #     ),
    # )
    # p.start()

    async def _runner():
        # Run the training in a new process
        return await _train_in_subprocess(
            model,
            train_loader,
            val_loader,
            test_loader,
            max_epochs,
            return_dict,
            queue,
        )

    task = asyncio.create_task(_runner())

    async def metric_generator(queue):
        # while p.is_alive():
        while not task.done():
            # Check if there are metrics in the queue
            try:
                metrics = queue.get_nowait()
                yield metrics
            except Exception:
                await asyncio.sleep(0.1)

    metrics_generator = metric_generator(queue)
    async for metrics in metrics_generator:
        yield metrics
    # p.join()

    yield return_dict["model"], return_dict["results"]


async def predict(
    model: models.MPNN,
    smis: List[str],
    in_new_thread: bool = True,
) -> np.ndarray:
    def _pred(model, smis):
        test_data = [data.MoleculeDatapoint.from_smi(smi) for smi in smis]
        featurizer = featurizers.SimpleMoleculeMolGraphFeaturizer()
        test_dset = data.MoleculeDataset(test_data, featurizer=featurizer)
        test_loader = data.build_dataloader(test_dset, shuffle=False)
        with torch.inference_mode():
            trainer = pl.Trainer(
                logger=False,
                enable_progress_bar=True,
                accelerator="auto",
                enable_checkpointing=False,
                enable_model_summary=False,
            )
            test_preds = trainer.predict(model, test_loader)
        test_preds = np.concatenate(test_preds, axis=0)

        return test_preds

    if in_new_thread:
        # Offload the blocking _pred function to a separate thread.
        return await asyncio.to_thread(_pred, model, smis)
    else:
        # Run _pred directly in the current coroutine.
        return _pred(model, smis)



from pytest_funcnodes import nodetest, all_nodes_tested
import funcnodes_chemprop as fnmodule  # noqa
import pandas as pd
import os
import numpy as np


def test_all_nodes_tested(all_nodes):
    all_nodes_tested(all_nodes, fnmodule.NODE_SHELF, ignore=[])


@nodetest(fnmodule.ChempropNode)
async def test_chemprop_node():
    # Test the ChempropNode with a dummy input
    df = pd.read_csv(os.path.join(os.path.dirname(__file__), "mol.csv"))
    smiles_column = "smiles"
    target_columns = "lipo"

    node = fnmodule.ChempropNode()

    node["df"] < df
    node["smiles_column"] < smiles_column
    node["target_columns"] < target_columns
    node["max_epochs"] < 2

    await node

    outdf = node["out_df"].value
    assert isinstance(outdf, pd.DataFrame), "Output should be a DataFrame"

    assert len(outdf) == len(df), (
        "Output DataFrame should have the same length as input"
    )

    assert "chemprop_role" in outdf.columns, (
        "Output DataFrame should contain 'chemprop_role' column"
    )

    assert "chemprop_prediction_lipo" in outdf.columns, (
        "Output DataFrame should contain 'chemprop_prediction' column"
    )

    # assert "chemprop_prediction_lipo" is all flaot

    assert np.issubdtype(outdf["chemprop_prediction_lipo"].dtype, np.floating), (
        "chemprop_prediction_lipo should be of float type"
    )

    # assert no nan

    assert not outdf["chemprop_prediction_lipo"].isnull().any(), (
        "NaN values found in chemprop_prediction_lipo"
    )


@nodetest(
    [fnmodule.make_model, fnmodule.MakeDataNode, fnmodule.predict, fnmodule.TrainNode]
)
async def test_make_model_node():
    # Test the ChempropNode with a dummy input
    df = pd.read_csv(os.path.join(os.path.dirname(__file__), "mol.csv"))
    smiles_column = "smiles"
    target_columns = "lipo"

    datanode = fnmodule.MakeDataNode()
    datanode["df"] < df
    datanode["smiles_column"] < smiles_column
    datanode["target_columns"] < target_columns

    await datanode

    modelnode = fnmodule.make_model()
    modelnode["scaler"] > datanode["scaler"]

    await modelnode

    model = modelnode["model"].value
    assert isinstance(model, fnmodule.models.MPNN), "Output should be a MPNN model"

    trainnode = fnmodule.TrainNode()
    trainnode["model"] > modelnode["model"]
    trainnode["train_loader"] > datanode["train_loader"]
    trainnode["val_loader"] > datanode["val_loader"]
    trainnode["test_loader"] > datanode["test_loader"]

    trainnode["max_epochs"] < 2

    await trainnode

    prednode = fnmodule.predict()

    prednode["model"] > trainnode["trained_model"]
    prednode["smiles"] < "CCC"

    await prednode

    preds = prednode["preds"].value
    assert isinstance(preds, np.ndarray), "Output should be a numpy array"
    assert preds.shape == (1, 1), "Output should have shape (1,)"



from funcnodes_chemprop import functions as fn
import pandas as pd
import os
from chemprop import models
import pytest


@pytest.mark.asyncio
async def test_chemprop_workflow():
    df = pd.read_csv(os.path.join(os.path.dirname(__file__), "mol.csv"))
    smiles_column = "smiles"
    target_columns = ["lipo"]

    (
        train_loader,
        val_loader,
        test_loader,
        scaler,
        train_indices,
        val_indices,
        test_indices,
    ) = await fn.make_data(
        df, smiles_column, target_columns, split=(0.8, 0.1, 0.1), in_new_thread=True
    )

    model = fn.make_model(scaler)
    inipreds = await fn.predict(model, df[smiles_column], in_new_thread=False)

    initerror = ((df[target_columns].values - inipreds) ** 2).mean() ** 0.5

    assert initerror > 1, "Initial RMSE should be greater than 1"

    assert isinstance(
        model,
        models.MPNN,
    )

    import time

    start_time = time.time()
    async for metrics in fn.train_in_subprocess(
        model, train_loader, val_loader, test_loader, max_epochs=20
    ):
        print("\n", time.time() - start_time)

    model, results = metrics
    preds = await fn.predict(model, df[smiles_column], in_new_thread=False)

    # rmse
    error = ((df[target_columns].values - preds) ** 2).mean() ** 0.5

    assert error < initerror, "RMSE should be less than initial RMSE"



import pandas as pd
from typing import List, Tuple, Union
from chemprop import models
from chemprop.data import SplitType
from sklearn.preprocessing import StandardScaler

from torch.utils.data import DataLoader
import numpy as np
import funcnodes as fn
from . import functions as f
import copy
import asyncio


@fn.NodeDecorator(
    id="chemprop.make_model",
    name="Make Model",
    description="Create a Chemprop model",
    outputs=[{"name": "model"}],
)
def make_model(scaler: StandardScaler) -> models.MPNN:
    mpnn = f.make_model(scaler)
    return mpnn


class MakeDataNode(fn.Node):
    node_id = "chemprop.make_data"
    node_name = "Make Data"
    description = "Create Chemprop data loaders"

    df = fn.NodeInput(uuid="df", type=pd.DataFrame)
    smiles_column = fn.NodeInput(uuid="smiles_column", type=str)
    target_columns = fn.NodeInput(uuid="target_columns", type=str)
    split = fn.NodeInput(
        uuid="split", type=Tuple[float, float, float], default=(0.8, 0.1, 0.1)
    )

    train_loader = fn.NodeOutput(uuid="train_loader", type=DataLoader)
    val_loader = fn.NodeOutput(uuid="val_loader", type=DataLoader)
    test_loader = fn.NodeOutput(uuid="test_loader", type=DataLoader)
    scaler = fn.NodeOutput(uuid="scaler", type=StandardScaler)

    def __init__(self):
        super().__init__()
        self.get_input("df").on("after_set_value", self._update_columns)

    def _update_columns(self, **kwargs):
        try:
            df = self.get_input("df").value
            smiles_column = self.get_input("smiles_column")
            target_columns = self.get_input("target_columns")
        except KeyError:
            return
        try:
            smiles_column.update_value_options(options=list(df.columns))
            target_columns.update_value_options(options=list(df.columns))
        except Exception:
            smiles_column.update_value_options(options=[])
            target_columns.update_value_options(options=[])

    async def func(self, df, smiles_column, target_columns, split):
        (
            train_loader,
            val_loader,
            test_loader,
            scaler,
            train_indices,
            val_indices,
            test_indices,
        ) = await f.make_data(df, smiles_column, target_columns, split)

        self.outputs["train_loader"].value = train_loader
        self.outputs["val_loader"].value = val_loader
        self.outputs["test_loader"].value = test_loader
        self.outputs["scaler"].value = scaler


# @fn.NodeDecorator(
#     id="chemprop.train",
#     name="Train",
#     description="Train a Chemprop model",
#     outputs=[{"name": "trained model"}, {"name": "results"}],
#     default_io_options={
#         "model": {"does_trigger": False},
#         "train_loader": {"does_trigger": False},
#         "val_loader": {"does_trigger": False},
#         "test_loader": {"does_trigger": False},
#         "max_epochs": {"does_trigger": False},
#     },
# )
# async def train(
#     model: models.MPNN,
#     train_loader: DataLoader,
#     val_loader: DataLoader,
#     test_loader: DataLoader,
#     max_epochs: int = 20,
# ) -> Tuple[models.MPNN, dict, pd.DataFrame]:
#     model_copy = copy.deepcopy(model)
#     metrics_df = None

#     async for metrics in f.train_in_subprocess(
#         model_copy, train_loader, val_loader, test_loader, max_epochs
#     ):
#         if isinstance(metrics, dict):
#             if metrics_df is None:
#                 metrics_df = pd.DataFrame([metrics])
#             else:
#                 metrics_df = pd.concat([metrics_df, pd.DataFrame([metrics])])
#         print(f"Received metrics: {metrics}")

#     return (*metrics, metrics_df)


class TrainNode(fn.Node):
    node_id = "chemprop.train"
    node_name = "Train"
    description = "Train a Chemprop model"

    model = fn.NodeInput(uuid="model", type=models.MPNN, does_trigger=False)

    train_loader = fn.NodeInput(
        uuid="train_loader", type=DataLoader, does_trigger=False
    )
    val_loader = fn.NodeInput(uuid="val_loader", type=DataLoader, does_trigger=False)
    test_loader = fn.NodeInput(uuid="test_loader", type=DataLoader, does_trigger=False)
    max_epochs = fn.NodeInput(
        uuid="max_epochs", type=int, default=20, does_trigger=False
    )

    trained_model = fn.NodeOutput(uuid="trained_model", type=models.MPNN)
    results = fn.NodeOutput(uuid="results", type=dict)
    metrics = fn.NodeOutput(uuid="metrics", type=pd.DataFrame)

    async def func(self, model, train_loader, val_loader, test_loader, max_epochs):
        model_copy = copy.deepcopy(model)

        metrics_df = None
        collected_metrics = []
        running = True

        async def metrics_publisher():
            last_published = 0
            while running:
                if len(collected_metrics) > last_published:
                    self.outputs["metrics"].value = pd.DataFrame(collected_metrics)
                    last_published = len(collected_metrics)
                await asyncio.sleep(1)

        publish_task = asyncio.create_task(metrics_publisher())

        with self.progress(
            desc="Training",
            total=max_epochs,
            unit="epoch",
            initial=0,
        ) as progress:
            async for metrics in f.train_in_subprocess(
                model_copy, train_loader, val_loader, test_loader, max_epochs
            ):
                if isinstance(metrics, dict):
                    ep = metrics["epoch"]
                    progress.update(int(ep - progress.n))
                    del metrics["epoch"]
                    converted_metrics = [
                        {"epoch": ep, "type": k, "value": v} for k, v in metrics.items()
                    ]
                    collected_metrics.extend(converted_metrics)

            running = False
            await publish_task
            # publish_thread.join()
        trained_model, results = metrics
        self.outputs["trained_model"].value = trained_model
        self.outputs["results"].value = results

        return trained_model, results, metrics_df


class ChempropNode(fn.Node):
    node_id = "chemprop.chemprop"
    node_name = "Chemprop"
    description = "Chemprop Node"

    model = fn.NodeInput(
        uuid="model", type=models.MPNN, does_trigger=False, default=None
    )
    df = fn.NodeInput(uuid="df", type=pd.DataFrame, does_trigger=False)
    smiles_column = fn.NodeInput(uuid="smiles_column", type=str, does_trigger=False)
    target_columns = fn.NodeInput(uuid="target_columns", type=str, does_trigger=False)
    split = fn.NodeInput(
        uuid="split",
        type=Tuple[float, float, float],
        default=(0.8, 0.1, 0.1),
        does_trigger=False,
        hidden=True,
    )
    split_type = fn.NodeInput(
        uuid="split_type",
        type=str,
        default="random",
        value_options={"options": [v.lower() for v in SplitType.values()]},
        does_trigger=False,
        hidden=True,
    )
    max_epochs = fn.NodeInput(
        uuid="max_epochs",
        type=int,
        default=20,
        does_trigger=False,
    )

    trained_model = fn.NodeOutput(uuid="trained_model", type=models.MPNN)
    results = fn.NodeOutput(uuid="results", type=dict)
    metrics = fn.NodeOutput(uuid="metrics", type=pd.DataFrame)
    out_df = fn.NodeOutput(uuid="out_df", type=pd.DataFrame)

    def __init__(self):
        super().__init__()
        self.get_input("df").on("after_set_value", self._update_columns)

    def _update_columns(self, **kwargs):
        try:
            df = self.get_input("df").value
            smiles_column = self.get_input("smiles_column")
            target_columns = self.get_input("target_columns")
        except KeyError:
            return
        try:
            smiles_column.update_value_options(options=list(df.columns))
            target_columns.update_value_options(options=list(df.columns))
        except Exception:
            smiles_column.update_value_options(options=[])
            target_columns.update_value_options(options=[])

    async def func(
        self,
        df,
        smiles_column,
        target_columns,
        model=None,
        max_epochs=20,
        split=(0.8, 0.1, 0.1),
        split_type="random",
    ):
        if isinstance(target_columns, str):
            target_columns = [target_columns]

        with self.progress(
            desc="Make Data",
        ) as progress:
            (
                train_loader,
                val_loader,
                test_loader,
                scaler,
                train_indices,
                val_indices,
                test_indices,
            ) = await f.make_data(
                df,
                smiles_column,
                target_columns,
                split,
                scaler=f.model_to_scaler(model) if model else None,
                split_type=split_type,
            )

        if model is None:
            model_copy = f.make_model(scaler)
        else:
            model_copy = copy.deepcopy(model)

        metrics_df = None
        collected_metrics = []
        running = True

        async def metrics_publisher():
            last_published = 0
            while running:
                if len(collected_metrics) > last_published:
                    self.outputs["metrics"].value = pd.DataFrame(collected_metrics)
                    last_published = len(collected_metrics)
                await asyncio.sleep(1)

        publish_task = asyncio.create_task(metrics_publisher())

        with self.progress(
            desc="Training",
            total=max_epochs,
            unit="epoch",
            initial=0,
        ) as progress:
            async for metrics in f.train_in_subprocess(
                model_copy, train_loader, val_loader, test_loader, max_epochs
            ):
                if isinstance(metrics, dict):
                    ep = metrics["epoch"]
                    progress.update(int(ep - progress.n))
                    del metrics["epoch"]
                    converted_metrics = [
                        {"epoch": ep, "type": k, "value": v} for k, v in metrics.items()
                    ]
                    collected_metrics.extend(converted_metrics)

            running = False
            await publish_task
            # publish_thread.join()
        trained_model, results = metrics

        out_df = df.copy()
        out_df["chemprop_role"] = "train"
        out_df.loc[train_indices, "chemprop_role"] = "train"
        out_df.loc[val_indices, "chemprop_role"] = "val"
        out_df.loc[test_indices, "chemprop_role"] = "test"
        preds = await f.predict(trained_model, df[smiles_column])
        out_df[["chemprop_prediction_" + s for s in target_columns]] = preds
        self.outputs["results"].value = results
        self.outputs["out_df"].value = out_df
        self.outputs["trained_model"].value = trained_model
        return trained_model, results, metrics_df, out_df


@fn.NodeDecorator(
    id="chemprop.predict",
    name="Predict",
    description="Predict using a Chemprop model",
    outputs=[{"name": "preds"}],
)
async def predict(model: models.MPNN, smiles: Union[str, List[str]]) -> np.ndarray:
    if isinstance(smiles, str):
        smiles = [smiles]
    return await f.predict(model, smiles)


NODE_SHELF = fn.Shelf(
    name="Chemprop",
    description="Chemprop nodes",
    nodes=[ChempropNode, make_model, MakeDataNode, TrainNode, predict],
    subshelves=[],
)
